// AIvan.c 
// the AI of Ivan my compainion robot 
// top level c program file
// 
// by Les Hall
// started Thu Aug 18, 2016, 3:24 PM
// 



// include header files
#include <stdlib.h>  // utility functions
#include <stdio.h>  // the standard input / output header
#include <math.h>  // math functions
#include "AIvan.h"  // Ivan's header file



// define structures
// 
// define the meta data structure
struct metaData {
  int days, awake, asleep;
};
// 
// define the input/output audio and video stream buffers
struct input_output {
  int video[SIZE_X][SIZE_Y];
  float audio[AUDIO_BUFFER_SIZE];
};
// 
// define the cell structure
struct cell {
  float weights[7];  // work like neural net weights
  float sum;  // sum of weighted inputs and previous values
  float threshold;  // when sum exceeds this value, cell triggers
  float storage;  // stores output to act as latch (prevents ***dataCreep***)
  float output;  // stores the triggered output value for one cycle
};
// 
// define the block structure
struct shape {
  struct cell cells[SIZE_C];  // declare the cells in a block
};
// 
// define the array structure
struct cognitiveArray {
  struct shape shapes[SIZE_X][SIZE_Y][SIZE_Z];  // the array of cells
};



// declare functions
void switch_on(FILE *fp, struct metaData *md, struct cognitiveArray *ca);  // ON
void wake_up();  // wake Ivan up
void think(FILE *fp);  // Ivan thinks like Rodin's "Thinking Man"
int imagine();  // Ivan is daydreaming (dreaming while awake)
void go_to_sleep();  // tell Ivan to go to sleep
void dream();  // Ivan is dreaming
void switch_off(FILE *fp);  // switch Ivan off
void print_mode();  // print Ivan's mode to stdout
void get_input(struct input_output *is, struct cognitiveArray *ca);
void iterate(FILE *fp, struct cognitiveArray *ca, 
  struct input_output *input, struct input_output *output, 
  int mode); // compute one iteration of calculations
void get_output(struct input_output *os, struct cognitiveArray *ca);
void diaplay(struct input_output *s);
void init(struct cognitiveArray *ca);
int min(int a, int b);
int max(int a, int b);



// the top level function
int main(int argc, char *argv[]) {

  // declare variables
  struct metaData meta;  // declare the meta data structure
  struct cognitiveArray ca;  // declare the array of block
  struct input_output input;  // holds input video and audio
  struct input_output output;  // holds output video and audio
  FILE *fp;  // file pointer to Ivan's long term memory
  int mode = off; // Ivan's mode of operation, defined in AIvan.h
  int seed = RAND_MAX/2;  // random number seed


  // seed the random number generator
  srand(seed);  // need to seed the generator with something that changes from run to run

  // switch Ivan on - read in long term memory from disk file
  // and launch all of Ivan's cellular processes
  switch_on(fp, &meta, &ca);
  int dayCounter = meta.days;

  // keep running loop as long as Ivan's on cycle counter is positive
  while (dayCounter > 0) {

    // reset counters
    int wakingCounter = meta.awake;
    int sleepingCounter = meta.asleep;

    // Ivan wakes up for another day
    wake_up(); // send a wake-up signal to all of Ivan's cellular processes
    init(&ca);  // initialize variables
    
    // keep runing loop as long as Ivan's day cycle counter is positive
    while(wakingCounter  > 0) {

      mode = awake;  // set Ivan's mode to awake

      // Ivan's daytime mode
      if (imagine() ) {  // if Ivan is using his imagination

        mode |= dreaming;  // set the mode accordingy

      }

      
      // get the input screen
      get_input(&input, &ca); 

      // compute one iteration of AIvan's cognitive array
      iterate(fp, &ca, &input, &output,  mode); 

      // get the output screen
      get_output(&output, &ca);

      // display the input and outpu screens
      diaplay(&input);  // input screen
      diaplay(&output);  // output screen


      // tidy up chores
      --wakingCounter; // decrement the day cycle counter

    }

    
    // go to sleep Ivan
    go_to_sleep();  // any chores in the process of going to sleep go here

    // sleep as long as Ivan's night cycle counter is positive
    while(sleepingCounter > 0) {
  
      // determine Ivan's mode for this cycle of the night
      mode = asleep;  // set Ivan's mode to asleep
      mode |= dreaming; // dream big Ivan, dream big

      // only dream if the dream flag is on in Ivan's mode
      if (mode & dreaming) {

        dream();  // tell Ivan's cellular processses to dream

      }

      // decrement the on cycle counter
      --sleepingCounter;
    
    }
  
    // decrement the on cycle counter
    --dayCounter;
  }
      
  // display the input and outpu screens
  diaplay(&input);  // input screen
  diaplay(&output);  // output screen

  // switch Ivan off
  switch_off(fp);  

  // end of program
  exit(0);

}



// switch Ivan on
void switch_on(FILE *fp, struct metaData *md, struct cognitiveArray *ca) {

  // define local variables
  int days, awake, asleep;

  // read in Ivan's disk file (his long term memory)
  if ((fp = fopen(fileName, fileMode)) == NULL) {  // open file and check for error`

    printf("error opening file");

  } else {

    // scan in the meta data
    fscanf(fp, "%d %d %d\n", &days, &awake, &asleep);  
    
    // assign the scanned data
    md->days = days;
    md->awake = awake;
    md->asleep = asleep;

    // allocate memory for shapes array by declaring the size of it
    

  }
  
  // print switch on message
  printf("\n\n\n\nIvan switched on!  \n");

}



// wake Ivan up
void wake_up() {

  // print welcome message
  printf("\tgood morning!  \n");

}



// do things while awake
void think() {
  
  // write out Ivan's state of mind
  printf("\t\tIvan is thinking\n");

}



// determine if Ivan is exercising his imagination
int imagine() {

  return ( rand() < (RAND_MAX/4) ) ? 1 : 0;

}




// tell Ivan to go to sleep
void go_to_sleep() {

  // print going to sleep message
  printf("\tgood night!\n");

}



// Ivan dreams at night
void dream() {

  // print going to sleep message
  printf("\t\tIvan is dreaming\n");

}



// turn off Ivan
void switch_off(FILE *fp) {

  // final print message
  printf("Ivan is switching off\n");

  // close Ivan's long term memory
  if ( fclose(fp) == EOF ) {


  } else {

    printf("error:  error closing file\n");

  }

  // final print message
  printf("Ivan is switched off\n\n\n\n");

}



// print Ivan's mode
void print_mode(int mode) {

  // begin with intro text
  printf("Ivan's mode is");

  // select modes with separate conditionals
  if (mode < 0)
    printf(" error");
  if (mode & off)
    printf(" off");
  if (mode & awake)
    printf(" awake");
  if (mode & asleep)
    printf(" asleep");
  if (mode & dreaming)
    printf(" dreaming");
  if (mode & imagining)
    printf(" imagining");

  // end with a newline
  printf("\n");

}



// get the input data stream
void get_input(struct input_output *is, struct cognitiveArray *ca) {

  // loop thru the array and fill in some values
  for (int y = 0; y < SIZE_Y; ++y) {  // loop thru y

    // calculate the Y dimension sine value
    float yVal = y + (float)SIZE_Y / 2.0;
    float yCos = (float)cos(2.0*3.14159 * (double)yVal / (double)SIZE_Y);

    for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

      // calculate the X dimension sine value
      float xVal = x + (float)SIZE_X / 2.0;
      float xCos = (float)cos(2.0*3.14159 * (double)xVal / (double)SIZE_X);

      // multimply xVal times yVal to make a wavy display
      float val = xCos * yCos;

      // convert the value to a positve value of one color magnitued
      val = 128 * (val + 1.0);

      for(int c = 0; c < SIZE_C; ++c) {  // loop thru c
  
        // put the display value into both 
        // the output and the storage locations
        // of the cognitive array data structure
        ca->shapes[x][y][0].cells[c].output = val;  // output value
        ca->shapes[x][y][0].cells[c].storage = val;  // storage value

      }
  
    }

  }
  
  // create an audio stream
  for(int i = 0; i < AUDIO_BUFFER_SIZE; ++i) {

    is->audio[i] = (float)sin(2.0*3.14159 * (double)i / (double)AUDIO_BUFFER_SIZE);

  }

}



// compute one iteration of the array!
void iterate(FILE *fp, struct cognitiveArray *ca, 
  struct input_output *input, struct input_output *output, 
  int mode) {

  // print the message that Ivan is thinking
  printf("\t\tIvan is thikning\n");

  // apply the input data steam to Ivan's cognitive array
  get_input(input, ca);
  
  // run the calculations for one iteration
  // and do not shift the storage value to the output just yet
  // loop thru the array
  for(int z = 0; z < SIZE_Z; ++z) {  // loop thru z

    for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y

      for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

        for(int c = 0; c < SIZE_C; ++c) {  // loop thru c

         for(int w = 0; w < SIZE_W; ++w) {  // loop thru w

            // use a case statement to convert the w value to a neighboring direction
            int X = x;
            int Y = y;
            int Z = z;
            switch (w) {
              case 1:X = max(x-1, 0); break;
              case 2:X = min(x+1, SIZE_X); break;
              case 3:Y = max(y-1, 0); break;
              case 4:Y = min(y+1, SIZE_Y); break;
              case 5:Z = max(z-1, 0); break;
              case 6:Z = min(z+1, SIZE_Z); break;
              default: X = x, Y = y, Z = z;
             }

             // multiply the neighbor index provided by the case statement 
             // by the corresponding weight index and add it to this cell's sum
             ca->shapes[x][y][z].cells[c].sum += 
               ca->shapes[x][y][z].cells[c].weights[w] * 
               ca->shapes[X][Y][Z].cells[c].output;
  
          }

          // OK the sum value has been updated 
          // (including possible small sums from previous iterations)
          // so now see if the threshold has been reached and fire the neuron
          // if the threshold has been exceeded
          if ( ca->shapes[x][y][z].cells[c].sum > ca->shapes[x][y][z].cells[c].threshold ) {

            // copy the sum to the storage value
            ca->shapes[x][y][z].cells[c].storage = 
              ca->shapes[x][y][z].cells[c].sum;


            // erase th sum value because it has transferred to the storage value
            ca->shapes[x][y][z].cells[c].sum = 0.0;


           // for now just let the fired value sit in storage.  
           // it will get transferred to the output in the next loop set
           // this way we don't have data creeping in every direction

          }

        }

      }

    }

  }

  // now shift all of the storage values to the output values
  // loop thru the array
  for(int z = 0; z < SIZE_Z; ++z) {  // loop thru z

    for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y

      for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

        for(int c = 0; c < SIZE_C; ++c) {  // loop thru c

          ca->shapes[x][y][z].cells[c].output = 
            ca->shapes[x][y][z].cells[c].storage;

        }

      }

    }

  }

  // apply the output data steam from Ivan's cognitive array
  get_output(output, ca);

}



// get the output data stream
void get_output(struct input_output *os, struct cognitiveArray *ca) {

  // and finally update the output values
  // deliver video output patterns to output structure
  // loop thru the array
  for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y
  
    for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

      os->video[x][y] = ca->shapes[x][y][SIZE_Z-1].cells[6].output;  

    }

  }
  // 
  // apply audio output buffer
  for(int i = 0; i < AUDIO_BUFFER_SIZE; ++i) {

    // just a null output for now until I figure out the mapping
    // between the cognitive array and the audio buffers 
    os->audio[i] = 0; 

  }
  
}



// display a video stream
void diaplay(struct input_output *s) {

  // determine the maximum brightness for scaling
  float max = 0;

  // loop thru the array
  for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y
  
    for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

      float val = s->video[x][y];
      if (val > max)
        max = val;

    }

  }

  printf("\n\n");

  // loop thru the array
  for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y
  
    for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

      int d = (int) ( 8 * s->video[x][y] / max );
      printf("%c", '1' + d);

    }
    
    printf("\n");

  }

  printf("\n\n");

}


// initialize variables
void init(struct cognitiveArray *ca) {
  
  // initialize weights
  // loop thru the array
  for(int z = 0; z < SIZE_Z; ++z) {  // loop thru z

    for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y

      for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

        for(int c = 0; c < SIZE_C; ++c) {  // loop thru c

          for(int w = 0; w < SIZE_W; ++w) {  // loop thru w

          ca->shapes[x][y][z].cells[c].weights[w] = 2.0 * ((float)rand()) / ((float)RAND_MAX) - 1.0;

          }

        }

      }

    }

  }

}



// define the min function
int min(int a, int b) {

  if (a >= b) {

    return b;

  } else {

    return a;

  }

}



// define the max function
int max(int a, int b) {

  if (a <= b) {

    return b;

  } else {

    return a;

  }

}


